from .substitution import Substitution
from pypika import Table


class AnalysisHelper:
    @staticmethod
    def table(name, alias):
        (schema, name) = name.split(".")
        return Table(name, schema=schema, alias=alias)

    @staticmethod
    def dimensions_from_names(dimensions, view):
        return list(filter(lambda x: x.name in dimensions, view.dimensions))

    # TODO: dedup this non dry situation
    @staticmethod
    def measures_from_names(measures, view):
        return list(filter(lambda x: x.name in measures, view.measures))

    @staticmethod
    def dimensions(dimensions, table):
        return [AnalysisHelper.field_from_dimension(d, table) for d in dimensions]

    @staticmethod
    def measures(measures, table):
        return [
            AnalysisHelper.field_from_measure(measure, table) for measure in measures
        ]

    @staticmethod
    def field_from_measure(measure, table):
        aggregate = Aggregate(measure, table)
        return aggregate.sql

    @staticmethod
    def field_from_dimension(d, table):
        sql = d.settings["sql"]
        substitution = Substitution(sql, table, dimension=d)
        return substitution.sql



import os
from dotenv import load_dotenv

load_dotenv()

# the values of those depend on your setup
MELTANO_POSTGRES_URL = os.getenv("MELTANO_POSTGRES_URL", "localhost")
MELTANO_POSTGRES_PORT = os.getenv("MELTANO_POSTGRES_PORT", "5432")
MELTANO_POSTGRES_USER = os.getenv("MELTANO_POSTGRES_USER", "meltano")
MELTANO_POSTGRES_PASSWORD = os.getenv("MELTANO_POSTGRES_PASSWORD", "meltano")
MELTANO_POSTGRES_DB = os.getenv("MELTANO_POSTGRES_DB", "meltano")
LOG_PATH = os.getenv("MELTANO_LOG_PATH", "meltano.log")
ENV = "development"

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

SQLALCHEMY_DATABASE_URI = f"postgresql+psycopg2://{MELTANO_POSTGRES_USER}:{MELTANO_POSTGRES_PASSWORD}@{MELTANO_POSTGRES_URL}:{MELTANO_POSTGRES_PORT}/{MELTANO_POSTGRES_DB}"

SQLALCHEMY_TRACK_MODIFICATIONS = False

# SQLALCHEMY_ECHO=True

THREADS_PER_PAGE = 2

SECRET_KEY = "damnitjanice"



import logging
import click
import sys

from meltano.core.permissions import grant_permissions, SpecLoadingError
from . import cli


@cli.group()
def permissions():
    """Database permission related commands."""
    pass


@permissions.command()
@click.argument("spec")
@click.option(
    "--db",
    help="The type of the target DB the specifications file is for.",
    type=click.Choice(["postgres", "snowflake"]),
    required=True,
)
@click.option("--dry", help="Do not actually run, just check.", is_flag=True)
def grant(db, spec, dry):
    """Grant the permissions provided in the provided specification file."""
    try:
        if not dry:
            click.secho("Error: Only dry runs are supported at the moment", fg="red")
            sys.exit(1)

        sql_commands = grant_permissions(db, spec, dry_run=dry)

        click.secho()
        click.secho("SQL Commands generated for given spec file:")

        for command in sql_commands:
            click.secho(f"{command};", fg="green")
            click.secho()
    except SpecLoadingError as exc:
        for line in str(exc).splitlines():
            click.secho(line, fg="red")
        sys.exit(1)



import logging
import re

from typing import Dict, List, Tuple, Set

from meltano.core.permissions.utils.snowflake_connector import SnowflakeConnector


GRANT_ROLE_TEMPLATE = "GRANT ROLE {role_name} TO {type} {entity_name}"

GRANT_PRIVILEGES_TEMPLATE = (
    "GRANT {privileges} ON {resource_type} {resource_name} TO ROLE {role}"
)

ALTER_USER_TEMPLATE = "ALTER USER {user_name} SET {privileges}"

GRANT_OWNERSHIP_TEMPLATE = (
    "GRANT OWNERSHIP"
    " ON {resource_type} {resource_name}"
    " TO ROLE {role_name} COPY CURRENT GRANTS"
)


def snowflaky(name: str) -> str:
    """
    Convert an entity name to an object identifier that will most probably be
    the proper name for Snoflake.

    e.g. gitlab-ci --> "gitlab-ci"
         527-INVESTIGATE$ISSUES.ANALYTICS.COUNTRY_CODES -->
         --> "527-INVESTIGATE$ISSUES".ANALYTICS.COUNTRY_CODES;

    Pronounced /snəʊfleɪkɪ/ like saying very fast snowflak[e and clarif]y
    Permission granted to use snowflaky as a verb.
    """
    name_parts = name.split(".")
    new_name_parts = []

    for part in name_parts:
        if re.match("^[0-9a-zA-Z_]*$", part) is None:
            new_name_parts.append(f'"{part}"')
        else:
            new_name_parts.append(part)

    return ".".join(new_name_parts)


def generate_grant_roles(entity_type: str, entity: str, config: str) -> List[str]:
    """
    Generate the GRANT statements for both roles and users.

    entity_type: "USER" or "ROLE"
    entity: the name of the entity (e.g. "yannis" or "REPORTER")
    config: the subtree for the entity as specified in the spec

    Returns the SQL commands generated as a list
    """
    sql_commands = []

    try:
        for member_role in config["member_of"]:
            sql_commands.append(
                GRANT_ROLE_TEMPLATE.format(
                    role_name=snowflaky(member_role),
                    type=entity_type,
                    entity_name=snowflaky(entity),
                )
            )
    except KeyError:
        logging.debug(
            "`member_of` not found for {}, skipping generation of GRANT ROLE statements.".format(
                entity
            )
        )

    return sql_commands


def generate_grant_privileges_to_role(
    role: str, config: str, shared_dbs: Set
) -> List[str]:
    """
    Generate all the privilege granting statements for a role.

    Most of the SQL command that will be generated are privileges granted to
    roles and this function orchestrates the whole process

    role: the name of the role (e.g. "LOADER" or "REPORTER") the privileges
          are GRANTed to
    config: the subtree for the role as specified in the spec
    shared_dbs: a set of all the shared databases defined in the spec.
                Used down the road by generate_database_grants() to also grant
                "IMPORTED PRIVILEGES" when access is granted to a shared DB.

    Returns the SQL commands generated as a list
    """
    sql_commands = []

    # Track all the DBs and schemas that have been given access to (GRANT USAGE)
    # the given role. Used in order to recursively grant the required access
    # to DBs or schemas implicetly reference when permissions are GRANTED for a
    # child Schema or Table.
    # Example: A role is given read access to table MY_DB.MY_SCHEMA.MY_TABLE
    # 1. In order to access MY_TABLE, the role has to be able to access MY_DB.MY_SCHEMA
    # 2. The script checks if USAGE on MY_DB has been granted to the role and
    #    assigns it to the role if not (and adds the DB to usage_granted["databases"])
    # 3. The same for the schema MY_SCHEMA
    # 4. Finaly the requested permissions are GRANTED to role for MY_TABLE
    usage_granted = {"databases": set(), "schemas": set()}

    try:
        for warehouse in config["warehouses"]:
            sql_commands.append(generate_warehouse_grants(role, warehouse))
    except KeyError:
        logging.debug(
            "`warehouses` not found for role {}, skipping generation of Warehouse GRANT statements.".format(
                role
            )
        )

    try:
        for database in config["privileges"]["databases"]["read"]:
            new_commands, usage_granted = generate_database_grants(
                role=role,
                database=database,
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.databases.read` not found for role {}, skipping generation of DATABASE read level GRANT statements.".format(
                role
            )
        )

    try:
        for database in config["privileges"]["databases"]["write"]:
            new_commands, usage_granted = generate_database_grants(
                role=role,
                database=database,
                grant_type="write",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.databases.write` not found for role {}, skipping generation of DATABASE write level GRANT statements.".format(
                role
            )
        )

    try:
        for schema in config["privileges"]["schemas"]["read"]:
            new_commands, usage_granted = generate_schema_grants(
                role=role,
                schema=schema,
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.schemas.read` not found for role {}, skipping generation of SCHEMA read level GRANT statements.".format(
                role
            )
        )

    try:
        for schema in config["privileges"]["schemas"]["write"]:
            new_commands, usage_granted = generate_schema_grants(
                role=role,
                schema=schema,
                grant_type="write",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.schemas.write` not found for role {}, skipping generation of SCHEMA write level GRANT statements.".format(
                role
            )
        )

    try:
        for table in config["privileges"]["tables"]["read"]:
            new_commands, usage_granted = generate_table_grants(
                role=role,
                table=table,
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.tables.read` not found for role {}, skipping generation of TABLE read level GRANT statements.".format(
                role
            )
        )

    try:
        for table in config["privileges"]["tables"]["write"]:
            new_commands, usage_granted = generate_table_grants(
                role=role,
                table=table,
                grant_type="write",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)
    except KeyError:
        logging.debug(
            "`privileges.tables.write` not found for role {}, skipping generation of TABLE write level GRANT statements.".format(
                role
            )
        )

    return sql_commands


def generate_warehouse_grants(role: str, warehouse: str) -> str:
    """
    Generate the GRANT statements for Warehouse usage (only one type at the moment).

    role: the name of the role the privileges are GRANTed to
    warehouse: the name of the warehouse (e.g. "transforming")

    Returns the SQL command generated
    """
    sql_grant = GRANT_PRIVILEGES_TEMPLATE.format(
        privileges="USAGE",
        resource_type="WAREHOUSE",
        resource_name=snowflaky(warehouse),
        role=snowflaky(role),
    )

    return sql_grant


def generate_database_grants(
    role: str, database: str, grant_type: str, usage_granted: Dict, shared_dbs: Set
) -> Tuple[List[str], Dict]:
    """
    Generate the GRANT statements for Databases.

    role: the name of the role the privileges are GRANTed to
    database: the name of the database (e.g. "RAW")
    grant_type: What type of privileges are granted? One of {"read", "write"}
    usage_granted: Passed by generate_grant_privileges_to_role() to track all
        all the entities a role has been granted access (USAGE) to.
    shared_dbs: a set of all the shared databases defined in the spec.

    Returns the SQL commands generated and the updated usage_granted as a Tuple
    """
    sql_commands = []

    usage_granted["databases"].add(database.upper())

    # If this is a shared database, we have to grant the "IMPORTED PRIVILEGES"
    #  privilege to the user and skip granting the specific permissions as
    #  "Granting individual privileges on imported databases is not allowed."
    if database in shared_dbs:
        sql_commands.append(
            GRANT_PRIVILEGES_TEMPLATE.format(
                privileges="IMPORTED PRIVILEGES",
                resource_type="DATABASE",
                resource_name=snowflaky(database),
                role=snowflaky(role),
            )
        )

        return (sql_commands, usage_granted)

    # And then grant privileges to the database
    if grant_type == "read":
        privileges = "USAGE"
    elif grant_type == "write":
        privileges = "USAGE, MONITOR, CREATE SCHEMA"
    else:
        raise SpecLoadingError(
            f"Wrong grant_type {spec_path} provided to generate_database_grants()"
        )

    sql_commands.append(
        GRANT_PRIVILEGES_TEMPLATE.format(
            privileges=privileges,
            resource_type="DATABASE",
            resource_name=snowflaky(database),
            role=snowflaky(role),
        )
    )

    return (sql_commands, usage_granted)


def generate_schema_grants(
    role: str, schema: str, grant_type: str, usage_granted: Dict, shared_dbs: Set
) -> Tuple[List[str], Dict]:
    """
    Generate the GRANT statements for SCHEMAs.

    role: the name of the role the privileges are GRANTed to
    schema: the name of the Schema (e.g. "RAW.PUBLIC")
    grant_type: What type of privileges are granted? One of {"read", "write"}
    usage_granted: Passed by generate_grant_privileges_to_role() to track all
        all the entities a role has been granted access (USAGE) to.
    shared_dbs: a set of all the shared databases defined in the spec.

    Returns the SQL commands generated and the updated usage_granted as a Tuple
    """
    sql_commands = []

    # Split the schema identifier into parts {DB_NAME}.{SCHEMA_NAME}
    #  so that we can check and use each one
    name_parts = schema.split(".")

    # Do nothing if this is a schema inside a shared database:
    #  "Granting individual privileges on imported databases is not allowed."
    if name_parts[0] in shared_dbs:
        return (sql_commands, usage_granted)

    if grant_type == "read":
        privileges = "USAGE"
    elif grant_type == "write":
        privileges = (
            "USAGE, MONITOR, CREATE TABLE,"
            " CREATE VIEW, CREATE STAGE, CREATE FILE FORMAT,"
            " CREATE SEQUENCE, CREATE FUNCTION, CREATE PIPE"
        )
    else:
        raise SpecLoadingError(
            f"Wrong grant_type {spec_path} provided to generate_schema_grants()"
        )

    # Before assigning privileges to a schema, check if
    #  usage to the database has been granted and
    # if not grant usage first to the database
    if name_parts[0].upper() not in usage_granted["databases"]:
        new_commands, usage_granted = generate_database_grants(
            role=role,
            database=name_parts[0],
            grant_type="read",
            usage_granted=usage_granted,
            shared_dbs=shared_dbs,
        )
        sql_commands.extend(new_commands)

    if name_parts[1] == "*":
        sql_commands.append(
            GRANT_PRIVILEGES_TEMPLATE.format(
                privileges=privileges,
                resource_type="ALL SCHEMAS IN DATABASE",
                resource_name=snowflaky(name_parts[0]),
                role=snowflaky(role),
            )
        )
    else:
        sql_commands.append(
            GRANT_PRIVILEGES_TEMPLATE.format(
                privileges=privileges,
                resource_type="SCHEMA",
                resource_name=snowflaky(schema),
                role=snowflaky(role),
            )
        )

    usage_granted["schemas"].add(schema.upper())

    return (sql_commands, usage_granted)


def generate_table_grants(
    role: str, table: str, grant_type: str, usage_granted: Dict, shared_dbs: Set
) -> Tuple[List[str], Dict]:
    """
    Generate the GRANT statements for TABLEs.

    role: the name of the role the privileges are GRANTed to
    table: the name of the TABLE (e.g. "RAW.PUBLIC.MY_TABLE")
    grant_type: What type of privileges are granted? One of {"read", "write"}
    usage_granted: Passed by generate_grant_privileges_to_role() to track all
        all the entities a role has been granted access (USAGE) to.
    shared_dbs: a set of all the shared databases defined in the spec.

    Returns the SQL commands generated and the updated usage_granted as a Tuple
    """
    sql_commands = []

    # Split the table identifier into parts {DB_NAME}.{SCHEMA_NAME}.{TABLE_NAME}
    #  so that we can check and use each one
    name_parts = table.split(".")

    # Do nothing if this is a table inside a shared database:
    #  "Granting individual privileges on imported databases is not allowed."
    if name_parts[0] in shared_dbs:
        return (sql_commands, usage_granted)

    if grant_type == "read":
        privileges = "SELECT"
    elif grant_type == "write":
        privileges = "SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES"
    else:
        raise SpecLoadingError(
            f"Wrong grant_type {spec_path} provided to generate_table_grants()"
        )

    # Generate the INFORMATION_SCHEMA identifier for that database
    #  in order to be able to filter it out
    info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

    # Before assigning privileges to a Table, check if
    #  usage to the database has been granted and
    # if not grant usage first to the database
    # Schemas will be checked as we generate them
    if name_parts[0].upper() not in usage_granted["databases"]:
        new_commands, usage_granted = generate_database_grants(
            role=role,
            database=name_parts[0],
            grant_type="read",
            usage_granted=usage_granted,
            shared_dbs=shared_dbs,
        )
        sql_commands.extend(new_commands)

    if name_parts[2] == "*":
        schemas = []

        if name_parts[1] == "*":
            # If {DB_NAME}.*.* was provided as the identifier, we have to fetch
            #  each schema in database DB_NAME, so that we can grant privileges
            #  for all tables in that schema
            # (You can not GRANT to all table with a wild card for the schema name)
            conn = SnowflakeConnector()
            db_schemas = conn.show_schemas(name_parts[0])
            for schema in db_schemas:
                if schema != info_schema:
                    schemas.append(schema)
        else:
            schemas = [f"{name_parts[0]}.{name_parts[1]}"]

        for schema in schemas:
            # first check if the role has been granted usage on the schema
            # either directly or indirectly (by granting to DB.*)
            if (
                schema.upper() not in usage_granted["schemas"]
                and f"{name_parts[0].upper()}.*" not in usage_granted["schemas"]
            ):
                new_commands, usage_granted = generate_schema_grants(
                    role=role,
                    schema=schema,
                    grant_type="read",
                    usage_granted=usage_granted,
                    shared_dbs=shared_dbs,
                )
                sql_commands.extend(new_commands)

            # And then grant privileges to all tables in that schema
            sql_commands.append(
                GRANT_PRIVILEGES_TEMPLATE.format(
                    privileges=privileges,
                    resource_type="ALL TABLES IN SCHEMA",
                    resource_name=snowflaky(schema),
                    role=snowflaky(role),
                )
            )
    else:
        # Check if the role has been granted usage on the schema
        schema = f"{name_parts[0]}.{name_parts[1]}"
        if (
            schema.upper() not in usage_granted["schemas"]
            and f"{name_parts[0].upper()}.*" not in usage_granted["schemas"]
        ):
            new_commands, usage_granted = generate_schema_grants(
                role=role,
                schema=schema,
                grant_type="read",
                usage_granted=usage_granted,
                shared_dbs=shared_dbs,
            )
            sql_commands.extend(new_commands)

        # And then grant privileges to the table
        sql_commands.append(
            GRANT_PRIVILEGES_TEMPLATE.format(
                privileges=privileges,
                resource_type="TABLE",
                resource_name=snowflaky(table),
                role=snowflaky(role),
            )
        )

    return (sql_commands, usage_granted)


def generate_alter_user(user: str, config: str) -> List[str]:
    """
    Generate the ALTER statements for USERs.

    user: the name of the USER
    config: the subtree for the user as specified in the spec

    Returns the SQL commands generated as a List
    """
    sql_commands = []
    alter_privileges = []

    if "can_login" in config:
        if config["can_login"]:
            alter_privileges.append("DISABLED = FALSE")
        else:
            alter_privileges.append("DISABLED = TRUE")

    if alter_privileges:
        sql_commands.append(
            ALTER_USER_TEMPLATE.format(
                user_name=snowflaky(user), privileges=", ".join(alter_privileges)
            )
        )

    return sql_commands


def generate_grant_ownership(role: str, config: str) -> List[str]:
    """
    Generate the GRANT OWNERSHIP statements for DATABASEs, SCHEMAs and TABLEs.

    role: the name of the role (e.g. "LOADER") OWNERSHIP will be GRANTed to
    config: the subtree for the role as specified in the spec

    Returns the SQL commands generated as a List
    """
    sql_commands = []

    try:
        for database in config["owns"]["databases"]:
            sql_commands.append(
                GRANT_OWNERSHIP_TEMPLATE.format(
                    resource_type="DATABASE",
                    resource_name=snowflaky(database),
                    role_name=snowflaky(role),
                )
            )
    except KeyError:
        logging.debug(
            "`owns.databases` not found for role {}, skipping generation of DATABASE OWNERSHIP statements.".format(
                role
            )
        )

    try:
        for schema in config["owns"]["schemas"]:
            name_parts = schema.split(".")
            info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

            if name_parts[1] == "*":
                schemas = []
                conn = SnowflakeConnector()
                db_schemas = conn.show_schemas(name_parts[0])

                for db_schema in db_schemas:
                    if db_schema != info_schema:
                        schemas.append(db_schema)

                for db_schema in schemas:
                    sql_commands.append(
                        GRANT_OWNERSHIP_TEMPLATE.format(
                            resource_type="SCHEMA",
                            resource_name=snowflaky(db_schema),
                            role_name=snowflaky(role),
                        )
                    )
            else:
                sql_commands.append(
                    GRANT_OWNERSHIP_TEMPLATE.format(
                        resource_type="SCHEMA",
                        resource_name=snowflaky(schema),
                        role_name=snowflaky(role),
                    )
                )
    except KeyError:
        logging.debug(
            "`owns.schemas` not found for role {}, skipping generation of SCHEMA OWNERSHIP statements.".format(
                role
            )
        )

    try:
        for table in config["owns"]["tables"]:
            name_parts = table.split(".")
            info_schema = f"{name_parts[0].upper()}.INFORMATION_SCHEMA"

            if name_parts[2] == "*":
                schemas = []

                if name_parts[1] == "*":
                    conn = SnowflakeConnector()
                    db_schemas = conn.show_schemas(name_parts[0])

                    for schema in db_schemas:
                        if schema != info_schema:
                            schemas.append(schema)
                else:
                    schemas = [f"{name_parts[0]}.{name_parts[1]}"]

                for schema in schemas:
                    sql_commands.append(
                        GRANT_OWNERSHIP_TEMPLATE.format(
                            resource_type="ALL TABLES IN SCHEMA",
                            resource_name=snowflaky(schema),
                            role_name=snowflaky(role),
                        )
                    )
            else:
                sql_commands.append(
                    GRANT_OWNERSHIP_TEMPLATE.format(
                        resource_type="TABLE",
                        resource_name=snowflaky(table),
                        role_name=snowflaky(role),
                    )
                )
    except KeyError:
        logging.debug(
            "`owns.tables` not found for role {}, skipping generation of TABLE OWNERSHIP statements.".format(
                role
            )
        )

    return sql_commands



from .base import Base
from ..app import db
from sqlalchemy.ext.mutable import MutableDict


class Settings(Base):
    __tablename__ = "setting"
    settings = db.Column(
        MutableDict.as_mutable(db.JSON()), nullable=False, default={"connections": []}
    )

    def serializable(self):
        return {"settings": self.settings}

    def __repr__(self):
        return f"<Settings id: {self.id}>"



import logging
import re

import sqlalchemy
from flask import jsonify
from pypika import Query, Order

from .analysishelper import AnalysisHelper
from .aggregate import Aggregate
from .date import Date
from .joinhelper import JoinHelper
from ..app import app, db
from ..models.data import View, DimensionGroup, Join
from ..models.projects import Project
from ..models.settings import Settings


class SqlHelper:
    def parse_sql(self, input):
        placeholders = self.placeholder_match(input)

    def placeholder_match(self, input):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input)
        inner_results = re.findall(inner_pattern, input)
        return (outer_results, inner_results)

    def get_names(self, things):
        return [thing.name for thing in things]

    def get_sql(self, explore, incoming_json):
        view_name = incoming_json["view"]
        view = View.query.filter(View.name == view_name).first()
        base_table = view.settings["sql_table_name"]
        (schema, table) = base_table.split(".")
        incoming_dimensions = incoming_json["dimensions"]
        incoming_dimension_groups = incoming_json["dimension_groups"]
        incoming_measures = incoming_json["measures"]
        incoming_filters = incoming_json["filters"]
        incoming_joins = incoming_json["joins"]
        incoming_limit = incoming_json.get("limit", 50)
        incoming_order = incoming_json["order"]
        order = None
        if incoming_order:
            if incoming_order["direction"] == "asc":
                order = Order.asc
            else:
                order = Order.desc
        orderby = incoming_order["column"] if incoming_order else None

        # get all timeframes
        timeframes = [t["timeframes"] for t in incoming_dimension_groups]
        # flatten list of timeframes
        timeframes = [y for x in timeframes for y in x]
        dimensions = AnalysisHelper.dimensions_from_names(incoming_dimensions, view)
        measures = AnalysisHelper.measures_from_names(incoming_measures, view)
        dimensions_raw = dimensions
        measures_raw = measures

        table = AnalysisHelper.table(base_table, explore.name)
        joins = [JoinHelper.get_join(j) for j in incoming_joins]
        dimension_groups = self.dimension_groups(
            view_name, incoming_dimension_groups, table
        )
        dimensions = AnalysisHelper.dimensions(dimensions, table)
        dimensions = dimensions + dimension_groups
        measures = AnalysisHelper.measures(measures, table)

        if orderby:
            ordered_by_column = [d for d in dimensions_raw if d.name == orderby]
            if ordered_by_column:
                orderby = self.dimensions(ordered_by_column, table)[0]
            else:
                ordered_by_column = [m for m in measures_raw if m.name == orderby]
                if ordered_by_column:
                    orderby = self.measures(ordered_by_column, table)[0]
                else:
                    raise Exception(
                        "Something is wrong, no dimension or measure column matching the column to sort by."
                    )

        column_headers = self.column_headers(dimensions_raw, measures_raw)
        names = self.get_names(dimensions_raw + measures_raw)
        return {
            "dimensions": dimensions_raw,
            "measures": measures_raw,
            "column_headers": column_headers,
            "names": names,
            "sql": self.get_query(
                from_=table,
                dimensions=dimensions,
                measures=measures,
                limit=incoming_limit,
                joins=joins,
                order=order,
                orderby=orderby,
            ),
        }

    def column_headers(self, dimensions, measures):
        return [d.label for d in dimensions + measures]

    def dimension_groups(self, view_name, dimension_groups, table):
        fields = []
        for dimension_group in dimension_groups:
            dimension_group_queried = (
                DimensionGroup.query.join(View, DimensionGroup.view_id == View.id)
                .filter(View.name == view_name)
                .filter(DimensionGroup.name == dimension_group["name"])
                .first()
            )
            (_table, name) = dimension_group_queried.table_column_name.split(".")
            for timeframe in dimension_group["timeframes"]:
                d = Date(timeframe, table, name)
                fields.append(d.sql)
        return fields

    def get_query(
        self, from_, dimensions, measures, limit, joins=None, order=None, orderby=None
    ):
        select = dimensions + measures
        q = Query.from_(from_)
        if joins:
            region = from_
            entry = joins[0]["table"]
            join_dimensions = joins[0]["dimensions"]
            join_measures = joins[0]["measures"]
            select = select + join_dimensions + join_measures
            dimensions = dimensions + join_dimensions
            q = q.join(joins[0]["table"]).on(
                region.id == entry.region_id and region.id == entry.region_id
            )
        q = q.select(*select).groupby(*dimensions)
        if order:
            q = q.orderby(orderby, order=order)
        q = q.limit(limit)
        return f"{str(q)};"

    def reset_db(self):
        try:
            Settings.__table__.drop(db.engine)
            Project.__table__.drop(db.engine)
            db.drop_all()
        except sqlalchemy.exc.OperationalError as err:
            logging.error("Failed drop database.")
        db.create_all()
        settings = Settings()
        db.session.add(settings)
        db.session.commit()
        return jsonify({"dropped_it": "like_its_hot"})



from enum import Enum

from .analysishelper import AnalysisHelper
from ..models.data import View, Join


class JoinType(Enum):
    left_join = "left"
    inner_join = "inner"
    full_outer_join = "full_outer"
    cross_join = "cross"


class JoinHelper:
    @staticmethod
    def get_join(join):
        view = JoinHelper.get_view(join)
        table = AnalysisHelper.table(view.settings["sql_table_name"], view.name)
        dimensions = JoinHelper.get_dimensions(join, view, table)
        # TODO finish this implementation with measures
        measures = []
        join = Join.query.filter(Join.name == join["name"]).first()
        return {
            "view": view,
            "dimensions": dimensions,
            "measures": measures,
            "table": table,
            "on": join.settings["sql_on"],
            "join": join,
        }

    def get_dimensions(join, view, table):
        dimensions = AnalysisHelper.dimensions_from_names(join["dimensions"], view)
        return AnalysisHelper.dimensions(dimensions, table)

    def get_view(join):
        return View.query.filter(View.name == join["name"]).first()
